#include "clip.h"

BOOL InsideBound(const Point &, const CBound *, CBound::BoundType, UINT_32);
Point BoundIntercept(const Point &, const Point &, const CBound *, CBound::BoundType, UINT_32);

//==========================================================
// Clip
//
// Takes a polygon described as a list of points and clips
// it against the edges of the bounding volume.  The
// resulting polygon is returned in the same list.
// Any points that are culled away are deleted.
// New points are allocated as necessary.
//==========================================================
//VOID Clip(CListInt &listPoints, const CBound *pCBound)
VOID Clip(CListInt &points, const CBound *bound)
{
   CListInt clippedPoints, culledPoints;

   clippedPoints.Init();
   culledPoints.Init();

   // fprintf(stderr,"Clipping a contour with %d points\n", points.GetLength());
   // Clip agains x and y
   for (int dim = 0; dim < 2; dim ++)
   {
      // Clip agains lower and upper bound
      for (CBound::BoundType bbt = CBound::bbtMin; bbt <= CBound::bbtMax; bbt = (CBound::BoundType)(bbt + 1) )
      {
	 CListIterInt pi;
	 Point *ppt_last, *ppt, *ppt_new;
	 BOOL last_in, curr_in;


	 // fprintf(stderr,"Starting against bound with %d points\n", points.GetLength());
	 if (!points.PeekFirst())
	 {
	    // No points left in polygon
	    return;
	 }
	 
	 ppt_last = (Point*)points.PeekLast();
	 last_in = InsideBound(*ppt_last,bound,bbt,dim);

	 pi.Init(&points);
	 for (ppt = (Point*) pi.PeekFirst(); ppt != NULL; ppt = (Point*)pi.PeekNext())
	 {
	   curr_in = InsideBound(*ppt,bound,bbt,dim);

	    if (curr_in != last_in)
	    {
	       // The polygon crossed the bound, add the intersection point
	       ppt_new = new Point();
	       *ppt_new = BoundIntercept(*ppt_last, *ppt, bound, bbt, dim);
	       clippedPoints.InsertLast(ppt_new, 0);
	    }

	    if ( curr_in )
	    {
	       clippedPoints.InsertLast(ppt, 0);
	    }
	    else
	    {
	       culledPoints.InsertLast(ppt, 0);
	    }

	    ppt_last = ppt;
	    last_in = curr_in;
	 }

	 points.RemoveAll();
	 while ( ppt = (Point*)culledPoints.RemoveFirst() ) delete ppt;
	 while ( ppt = (Point*)clippedPoints.RemoveFirst() ) points.InsertLast(ppt, 0);

	 // fprintf(stderr,"Ending against bound with %d points\n", points.GetLength());
      }
   }
   // fprintf(stderr,"Drawing a contour with %d points\n", points.GetLength());
}

//==========================================================
// InsideBound
//
// returns TRUE if pt is inside the given edge of the bound.
//==========================================================
BOOL
InsideBound(const Point &pt, const CBound *bound, CBound::BoundType bbt, UINT_32 dimension)
{
   ASSERT( bbt < CBound::bbtSize );
   ASSERT( dimension >= 0 && dimension < 4 );

   if ( bbt == CBound::bbtMin ) 
     {
       // fprintf(stderr, "Comparing %f >= %f?\n", pt(dimension), Get(bbtMin)(dimension));
       return ( pt(dimension) >= bound->Get(CBound::bbtMin)(dimension) );
     }
   else 
     {
       // fprintf(stderr, "Comparing %f <= %f?\n", pt(dimension), Get(bbtMin)(dimension));
       return ( pt(dimension) <= bound->Get(CBound::bbtMax)(dimension) );
     }
}

//==========================================================
// BoundIntercept
//
// returns the intersection of the line segment pt1-pt2
// with the given edge of the bound.
//==========================================================
Point
BoundIntercept(const Point &pt1, 
	       const Point &pt2, 
	       const CBound *bound, 
	       CBound::BoundType bbt, 
	       UINT_32 dimension)
{
   ASSERT( bbt < CBound::bbtSize );
   ASSERT( dimension >= 0 && dimension < 4 );
   ASSERT( InsideBound(pt1,bound,bbt,dimension) != InsideBound(pt2,bound,bbt,dimension) );

   FLOAT t;

   t = ( bound->Get(bbt)(dimension) - pt1(dimension) ) / ( pt2(dimension) - pt1(dimension) );
   return ( pt1 + t * (pt2 - pt1) );
}
