#ifndef __VERTEX_H__
#define __VERTEX_H__

#include "vector.h"
#include "point.h"
class CSLIDESurface;

//////////////////////////////////////////////////
// Operations Supported
//
// Vertex p, q, r;
// Vector u, v, w;   // Vectors
// FLOAT f;          // Scalars
// BOOL b;           // Booleans
//
// f = p[0];       // LHS Index Operation
// p[0] = f;       // LHS Index Operation
// f = p(0);       // RHS Index Operation
// p(0) = f;       // ILLEGAL -- RHS (right hand side) only!!!
//
// p = q;
// p = v;
// p += v;
// p -= v;
// p = q + v;
// v = p - q;
// f = p.Dot(v);   // Dot Product for on Plane calculations

// Operations for Averaging to find Centroids and Doing Linear Combinations
// p += q;
// p *= f;         // Scalar Multiplication
// p /= f;         // Scalar Multiplication
// p = q + r;
// p = q * f;      // Scalar Multiplication
// p = f * q;      // Scalar Multiplication
// p = q / f;      // Scalar Division
// p = f / q;      // Scalar Inversion
//

// Logic
// b = (p == q); // Quasi Equal allows for floating vertex inaccuracy
// b = (p != q); // Quasi Not Equal allows for floating vertex inaccuracy

// Bounding Box Checking Operations
// b = (p < q);  // Elementwise
// b = (p <= q); // Elementwise
// b = (p > q);  // Elementwise
// b = (p >= q); // Elementwise

//////////////////////////////////////////////////////////////////////
// Vertex Class
//

class Vertex
{
public:
  Vertex();
  Vertex(const Point &pt, const Vector &vNormal, const Vector &vColor, 
	 const Point &ptTexture);
  ~Vertex();

public:
  VOID Null();
  BOOL Homogenize();

public:
  Vertex &operator=(const Vertex &vx);
  Vertex &operator+=(const Vertex &vx);
  Vertex &operator-=(const Vertex &vx);
  Vertex &operator*=(const FLOAT f);
  Vertex &operator/=(const FLOAT f);

public:
//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 8: Gouraud Shading
//
  VOID Render();
//
// END  : ASSIGNMENT 8: Gouraud Shading
//////////////////////////////////////////////////////////////////////////////

public:
  inline Point &SetPoint();
  inline Vector &SetNormal();
  inline Vector &SetColor();
  inline Point &SetTexture();
  inline const Point &GetPoint() const;
  inline const Vector &GetNormal() const;
  inline const Vector &GetColor() const;
  inline const Point &GetTexture() const;

  inline CSLIDESurface *GetSurface();
  inline VOID SetSurface(CSLIDESurface *pCSLIDESurface);

protected:
  Point m_pt;
  Vector m_vNormal; // We will not be interpolating the normals
  Vector m_vColor;
  Point m_ptTexture;

  CSLIDESurface *m_pCSLIDESurface;
};

//////////////////////////////////////////////////////////////////////
// External Operators
//

inline Vertex operator+(const Vertex &vxA, const Vertex &vxB);
inline Vertex operator-(const Vertex &vxA, const Vertex &vxB);

inline Vertex operator*(const Vertex &vx, const FLOAT f);
inline Vertex operator*(const FLOAT f, const Vertex &vx);
inline Vertex operator/(const Vertex &vx, const FLOAT f);

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline Point &Vertex::SetPoint()
{
  return m_pt;
}

inline Vector &Vertex::SetNormal()
{
  return m_vNormal;
}

inline Vector &Vertex::SetColor()
{
  return m_vColor;
}

inline Point &Vertex::SetTexture()
{
  return m_ptTexture;
}

inline const Point &Vertex::GetPoint() const
{
  return m_pt;
}

inline const Vector &Vertex::GetNormal() const
{
  return m_vNormal;
}

inline const Vector &Vertex::GetColor() const
{
  return m_vColor;
}

inline const Point &Vertex::GetTexture() const
{
  return m_ptTexture;
}

inline CSLIDESurface *Vertex::GetSurface()
{
  return m_pCSLIDESurface;
}

inline VOID Vertex::SetSurface(CSLIDESurface *pCSLIDESurface)
{
  m_pCSLIDESurface = pCSLIDESurface;
}

//////////////////////////////////////////////////////////////////////
// External Operator Implementations
//

inline Vertex operator+(const Vertex &vxA, const Vertex &vxB)
{
  return Vertex(vxA.GetPoint() + vxB.GetPoint(), 
		vxA.GetNormal() + vxB.GetNormal(), 
		vxA.GetColor() + vxB.GetColor(),
		vxA.GetTexture() + vxB.GetTexture());
}

inline Vertex operator-(const Vertex &vxA, const Vertex &vxB)
{
  Vector v;
  Point pt, ptTexture;

  v = vxA.GetPoint() - vxB.GetPoint();
  pt.Set(v[0], v[1], v[2], v[3]);
  v = vxA.GetTexture() - vxB.GetTexture();
  ptTexture.Set(v[0], v[1], v[2], v[3]);

  return Vertex(pt,
		vxA.GetNormal() - vxB.GetNormal(), 
		vxA.GetColor() - vxB.GetColor(),
		ptTexture);
}

inline Vertex operator*(const Vertex &vx, const FLOAT f)
{
  return Vertex(vx.GetPoint() * f,
		vx.GetNormal() * f,
		vx.GetColor() * f,
		vx.GetTexture() * f);
}

inline Vertex operator*(const FLOAT f, const Vertex &vx)
{
  return Vertex(f * vx.GetPoint(),
		f * vx.GetNormal(),
		f * vx.GetColor(),
		f * vx.GetTexture());
}

inline Vertex operator/(const Vertex &vx, const FLOAT f)
{
  ASSERT( f != 0.0 );

  return Vertex(vx.GetPoint() / f,
		vx.GetNormal() / f,
		vx.GetColor() / f,
		vx.GetTexture() / f);
}

#endif // __VERTEX_H__

